importScripts('tf.min.js');
importScripts('tf-backend-wasm.min.js');

tf.enableProdMode();

let nsfwModel;

const SIZE = 224;
const NSFW_CLASSES = {
  0: 'Drawing',
  1: 'Hentai',
  2: 'Neutral',
  3: 'Porn',
  4: 'Sexy'
};

function customFetchFunc(url, o) {
  if (url === 'models/mobilenet_v2/group1-shard1of1') {
    return self.fetch(`${url}.bin`, o);
  }
  return self.fetch(url, o);
}

function nsfwProcess(values) {
  const topK = 5;
  const result = {};
  const valuesAndIndices = [];
  const topkValues = new Float32Array(topK);
  const topkIndices = new Int32Array(topK);

  for (let i = 0; i < values.length; i++) {
    valuesAndIndices.push({ value: values[i], index: i });
  }

  valuesAndIndices.sort((a, b) => b.value - a.value);

  for (let i = 0; i < topK; i++) {
    topkValues[i] = valuesAndIndices[i].value;
    topkIndices[i] = valuesAndIndices[i].index;
  }

  const resultArray = [];
  for (let i = 0; i < 5; i++) {
    result[NSFW_CLASSES[[topkIndices[i]]]] = Number.parseFloat((topkValues[i] * 100).toFixed(2));
    resultArray.push({ className: NSFW_CLASSES[[topkIndices[i]]], probability: Number.parseFloat((topkValues[i] * 100).toFixed(2)) });
  }

  return resultArray;
}

async function detectNSFW(bitmap) {
  const { width: w, height: h } = bitmap;
  const offScreen = new OffscreenCanvas(w, h);
  const ctx = offScreen.getContext('2d');
  ctx.drawImage(bitmap, 0, 0, w, h);

  const canvasData = ctx.getImageData(0, 0, w, h).data;
  const img = new ImageData(canvasData, w, h);
  const pixels = tf.browser.fromPixels(img);
  const normalized = pixels.toFloat().div(tf.scalar(255));

  let resized = normalized;
  if (pixels.shape[0] !== SIZE || pixels.shape[1] !== SIZE) {
    resized = tf.image.resizeBilinear(normalized, [SIZE, SIZE], true);
  }

  const batched = resized.reshape([1, SIZE, SIZE, 3]);
  const predictions = await nsfwModel.predict(batched);
  const values = await predictions.data();
  const result = nsfwProcess(values);
  predictions.dispose();
  self.postMessage({ result });
}

async function init({ data }) {
  // 初始化
  if (typeof data === 'string' && data === 'init') {
    await tf.setBackend('wasm');

    try {
      nsfwModel = await tf.loadLayersModel('indexeddb://nsfwModel');
      console.log('NSFW Model Loaded!');
    } catch (e) {
      try {
        nsfwModel = await tf.loadLayersModel('models/mobilenet_v2/model.json', { fetchFunc: customFetchFunc });
        await nsfwModel.save('indexeddb://nsfwModel');
        console.log('NSFW Model Saved!');
      } catch (e) {
        console.log('NSFW Model load fail!');
        console.error(e);
      }
    } finally {
      // warm up
      const result = tf.tidy(() => nsfwModel.predict(tf.zeros([1, SIZE, SIZE, 3])));
      await result.data();
      result.dispose();
    }
    self.postMessage({ init: true });
    return;
  }

  // 检查图片
  const bitmap = await createImageBitmap(data);
  detectNSFW(bitmap);
}

addEventListener('message', init);
